#include <cstdlib>
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>

#include <CLI/CLI.hpp>

#include <suplibxx/str.h>

#include "RdbStats.h"

#include "Options.h"
#include "config.h"
#include "usage.hpp"

Options::Options( int argc, char* argv[] ) {

    CLI::App app( "Compute statistics on an rdb table" );

    try {
        std::string group_str;
        std::string percentile_str;
        std::string override_str;

        app.add_option("-i,--input", input, "Input RDB file" )->check(CLI::ExistingFile);

        app.add_option("-o,--output", output, "Output RDB file" );

        auto opt_all = app.add_flag("-a,--all", all, "Generate statistics for all numerical columns" );

        auto opt_group = app.add_option( "-g,--group,-b,--break", group_str, "group data by the specified columns" )->join(',');
        auto opt_rows = app.add_option( "-r,--rows", rows, "ranges of rows to operate on" );

        opt_group->excludes( opt_rows );

        auto normalize_map = std::map<std::string,Normalize>( {
                                                  { "ave", Normalize::Average },
                                                  { "med", Normalize::Median  },
                                                  { "none", Normalize::None   }
            } );

        app.add_option( "-n,--normalize", normalize, "Normalize the output statistics" )
            ->transform( CLI::Transformer( normalize_map ));

        app.add_option("-p,--percentiles", percentile_str, "Generate percentile statistics" )->join(',');

        app.add_flag( "-q,--quartiles", quartiles, "Generate quartile statistics" );

        app.add_option("-d,--defn", override_str, "override column definition" )->join();

        app.add_flag( "--manual,--usage", manual, "output detailed help, then exit" );

        std::string include_exact_str;
        app.add_option("--include,-I", include_exact_str, "name of column to include" )->join();
        std::string include_re_str;
        app.add_option("--include-re,-J", include_re_str, "regular expression matching  column to include" )->join();

        std::string exclude_exact_str;
        app.add_option("--exclude,-X", exclude_exact_str, "name of column to exclude" )->join();
        std::string exclude_re_str;
        app.add_option("--exclude-re,-Y", exclude_re_str, "regular expression matching column to exclude" )->join();

        auto opt_columns = app.add_option( "columns", include_exact, "specific columns to process" );
        opt_all->excludes(opt_columns);

        bool debug = false;
        app.add_flag( "-D", debug, "debug" );

        bool version = false;
        app.add_flag( "--version,-v", version, "output version, then exit" );

        app.parse( argc, argv );

        // parse percentiles
        {
            using size_type = std::string::size_type;
            size_type length = percentile_str.size();
            size_type next;

            for( size_type first = 0 ; first < length ; first += next + 1) {
                std::string label{ percentile_str.substr( first, length - first  ) };
                double percentile;
                try {
                    percentile = std::stod( label, &next );
                }
                catch ( std::invalid_argument& e ) {
                    throw Exception( "error parsing percentile (" + label + ")" );
                }
                catch ( std::out_of_range& e ) {
                    throw Exception( "percentile (" + label + "): must be a real number within [ 0.0, 100.0 ]" );
                }

                if ( next != length - first && label[next] != ',' )
                    throw Exception( "error parsing percentile (" + label + "): unexpected character" + label[next] );
                if ( percentile < 0 || percentile > 100 )
                    throw Exception( "percentile (" + label + "): must be a real number within [ 0.0, 100.0 ]" );
                percentiles.emplace_back( "_p" + label.substr(0,next), percentile );
            }
        }

        // add quartiles if
        // 1. explicitly requested
        // 2. normalize by median requested and no percentiles specified
        if ( quartiles || ( normalize == Normalize::Median && percentiles.empty() ) ) {
            percentiles.emplace_back( "_fq", 25 );
            percentiles.emplace_back( "_lq", 75 );
        }

        // handle column type redefinitions
        {
            std::vector<std::string> override_defs;
            suplib::tok( override_defs, override_str, "\n" );

            for ( const auto& override_def : override_defs ) {
                std::vector<std::string> defspec;
                suplib::tok( defspec, override_def, "," );

                switch ( defspec.size() ) {

                case 1:
                    override_defn.emplace_back( defspec[0], OverrideType::Toggle );
                    break;

                case 2:
                    if      ( defspec[1] == "N" ) override_defn.emplace_back( defspec[0], OverrideType::Number );
                    else if ( defspec[1] == "S" ) override_defn.emplace_back( defspec[0], OverrideType::String );
                    else throw Exception( "Unknown definition override type: " + defspec[1] );
                    break;

                default:
                    throw Exception( "unable to parse definition override specification: " + override_str );
                    break;
                }
            }
        }

        suplib::tok( group, group_str, "," );

        std::vector<std::string> include_opt_exact;
        suplib::tok( include_opt_exact, include_exact_str, "\n" );
        std::copy( include_opt_exact.begin(), include_opt_exact.end(), std::back_inserter( include_exact ) );
        
        suplib::tok( include_re, include_re_str, "\n" );
        suplib::tok( exclude_exact, exclude_exact_str, "\n" );
        suplib::tok( exclude_re, exclude_re_str, "\n" );

        if ( debug )
            std::cerr << *this;

        if ( manual ) {
            usage();
            exit_requested = true;
            exit_value = EXIT_SUCCESS;
        }

        if ( version ) {
            std::cout << PACKAGE_STRING << "\n";
            exit_requested = true;
            exit_value = EXIT_SUCCESS;
        }
    }

    catch ( const CLI::ParseError &e ) {
        exit_requested = true;
        exit_value = app.exit(e);
    }

}

void
Options::print( std::ostream& os ) {

    os << "include columns: " << to_string( include_exact ) << "\n";
    os << "include columns (re): " << to_string( include_re ) << "\n";

    os << "exclude columns: " << to_string( exclude_exact ) << "\n";
    os << "exclude columns (re): " << to_string( exclude_re ) << "\n";

    os << "all: " << ( all ? "true" : "false" ) << "\n";

    os << "groups: " << to_string( group ) << "\n";

    os << "input: " << ( input.empty() ? "stdin" : input ) << "\n";
    os << "output: " <<( output.empty() ? "stdout" : output ) << "\n";

    os << "normalize: " << to_string( normalize ) << "\n";

    os << "override: " << to_string( override_defn ) << "\n";
    os << "percentiles: " << to_string( percentiles ) << "\n";

    os << "quartiles: " << (quartiles ? "true" : "false") << "\n";
    os << "rows: " << to_string( rows ) << "\n";
}

std::ostream& operator << ( std::ostream& os, Options& a ) {
    a.print( os );
    return os;
}
