// File:  rdbstats.cc

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rdbstats
//
// rdbstats is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rdbstats is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>

#include <rdbxx/RDB.h>
#include <Exception/Exception.h>

#include "Options.h"
#include "RdbStats.h"
#include "RdbStatsAve.h"
#include "RdbStatsPercentiles.h"
#include "RdbStatsPercentilesAve.h"
#include "RdbStatsPercentilesMed.h"
#include "Row.h"
#include "SelectedCols.h"
#include "StatsResult.h"
#include "config.h"

/**
   Use Rows or groups to select to calculate the statistics.
 */
template<class Type>
void rdbstats( RDB& irdbtable, Type& rdb_stats_columns,
	       const std::vector< std::pair< long, long > >& ranges, RDB& ordbtable, bool grouped ) {

    StatsResult< Type > stats_result( rdb_stats_columns, ordbtable );

    int status = RDB::REOL;

    int line_number = 1;
    for( const auto& range : ranges ) {

      // Skip the un-requested rows.
      // At the end of this loop, line_number == range[ ii ].first
        while( line_number < range.first &&
               RDB::REOF != (status = irdbtable.read( ))
               )
            ++line_number;

        if ( status == RDB::REOF )
            break;
               
      // loop through the requested range of rows.
      for ( ; line_number <= range.second; ++line_number ) {
	if ( RDB::REOF == (status = irdbtable.read( )) ) {
	  if ( ! grouped ) 
	    stats_result.calculate_statistics( ); 
	  return;
	} 

	stats_result.update_statistics( );

	if ( RDB::REOG & status ) 
	  stats_result.calculate_statistics( );
	//      	ordbtable.autoIdx( false );
      }
    }

    stats_result.calculate_statistics( );
}

/**
 * To decide what kind of statistics to be calculated!
 */
void rdbstats( Options& options ) {

    /*
     * get the struct of options from the parser class so that you can
     * access the option values.
     */
    // The input stream shall be the standard in
    RDB irdbtable(
                  [&](){ return 
                          options.input.empty()
                          ? RDB( &std::cin )
                          : RDB( options.input );
                  }()
                  );
    // The output stream shall be the standard out
    //    ostream* os = &cout;
    //    os->precision( 15 );
    RDB ordbtable(
                  [&](){ return
                          options.output.empty()
                          ? RDB( &std::cout )
                          : RDB( options.output, std::ios::out );
                  }()
                  );
    ordbtable.setComment( irdbtable );

    // if the all-columns option has been selected, clear any selected
    // on the command line
    if ( options.all )
        options.include_exact.clear();

    SelectedCols selected_cols( irdbtable,
                                options.include_exact,
                                options.include_re,
                                options.exclude_exact,
                                options.exclude_re,
                                options.group, options.override_defn );

    const std::vector< std::string >& colnames = selected_cols.get_selected_cols( );

    if ( colnames.empty() )
        throw Exception( "no columns were selected or specified" );

    // Set the group.
    bool grouped = ! options.group.empty();
    for ( const auto& colname : options.group ) {
	ordbtable.setColumn( irdbtable.getColumn( colname ) );
	irdbtable.setGroup( colname );
    }

    Row row( options.rows );

    if ( ! options.percentiles.empty() ) {

      switch( options.normalize ) {
      case Normalize::Average:
	{
	  std::vector< RdbStatsPercentilesAve > rdb_stats_columns;

	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname, options.percentiles );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      case Normalize::Median:
	{
	  std::vector< RdbStatsPercentilesMed > rdb_stats_columns;
	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname, options.percentiles );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      case Normalize::None:
	{
	  std::vector< RdbStatsPercentiles > rdb_stats_columns;
	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname, options.percentiles );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      }

    } else {

      switch( options.normalize ) {
      case Normalize::Average:
	{
	  std::vector< RdbStatsAve > rdb_stats_columns;
	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      case Normalize::Median:
	{
	  std::vector< RdbStatsPercentilesMed > rdb_stats_columns;
	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      case Normalize::None:
	{
	  std::vector< RdbStats > rdb_stats_columns;
	  for ( const auto& colname : colnames )
	    rdb_stats_columns.emplace_back( irdbtable, colname );
	  rdbstats( irdbtable, rdb_stats_columns, row.get_range( ), ordbtable, grouped );
	}
	break;
      }
    }
}

int main( int argc, char* argv[] ) {

#ifdef TRACEFCT
  TraceFct tf( argv[0], 1, -1 );
#endif

  try {

    Options options( argc, argv );

    if ( options.exit_requested )
        return options.exit_value;

    rdbstats( options );

    return EXIT_SUCCESS;

  } catch( Exception& E ) {

    std::cerr << E << std::endl;
    return EXIT_FAILURE;

  } catch( std::exception &e ) {

    std::cerr << e.what() << std::endl;
    return EXIT_FAILURE;

  } catch ( ... ) {

    std::cerr << "Unknown exception caught" << std::endl;

    return EXIT_FAILURE;
  }

  return 0;
}
